<?php

/**
 * @author Dominique Fournier <dominique@fournier38.fr>
 * @license BSD
 */

declare(strict_types=1);

namespace App\Service;

use React\ChildProcess\Process;

/**
 * This class manage the parallel ping
 */
class AsyncPing
{
    /**
     * @var array<string, bool>
     */
    private array $state = [];

    /**
     * Last chunk time
     * @var array<int>
     */
    private array $startTime = [];

    /**
     * @param array<string, array<string>> $resolved The dns=>ips to be pinged
     * @param callable $callable Method called on each line with the
     * PingStateInterface object as parameter
     */
    public function init(array $resolved, callable $callable): void
    {
        foreach ($resolved as $name => $ips) {
            foreach ($ips as $ip) {
                $this->startTime[$ip] = time();
                $callable(new PingStateNull([ "ip" => $ip , "startTime" => $this->startTime[$ip] ]));
            }
        }
    }

    /**
     * @param array<string, array<string>> $resolved The dns=>ips to be pinged
     * @param callable $callable Method called on each line with the
     * PingStateInterface object as parameter
     */
    public function run(array $resolved, callable $callable): void
    {
        if (PHP_VERSION_ID < 80000) {
            exit('Socket descriptors require PHP 8+' . PHP_EOL);
        }
        $pingCommand = $this->pingCmdInPath();

        foreach ($resolved as $name => $ips) {
            foreach ($ips as $ip) {
                if (key_exists($ip, $this->state)) {
                    continue;
                }

                $this->startTime[$ip] = time();

                $cmd = "LANGUAGE=C $pingCommand -n -O -W 0.1 " . escapeshellarg($ip);
                $process = new Process(
                    $cmd,
                    null,
                    null,
                    [
                        ['socket'],
                        ['socket'],
                        ['socket']
                    ]
                );
                $this->state[$ip] = true;
                $process->start();

                if ($process->stdout !== null) {
                    $process->stdout->on('data', function ($chunk) use ($ip, $callable) {
                        preg_match_all(
                            "~(?P<bytes>\d+) bytes from (?P<ip>[0-9a-f:.]+).*: " .
                            "icmp_seq=(?P<seq>\d+) ttl=(?P<ttl>\d+) " .
                            "time=(?P<time>\d+.\d+) (?P<timeunit>.+)$~",
                            $chunk,
                            $matches
                        );
                        if (! empty($matches[0])) {
                            $matches["startTime"][0] = $this->startTime[$ip];
                            $matches = array_filter($matches, "is_string", ARRAY_FILTER_USE_KEY);
                            $matches = array_map("current", $matches);
                            $matches = array_filter($matches);
                            $this->startTime[$ip] = time();
                            $callable(new PingStateOK($matches));
                        }
                        preg_match_all("~(?P<noanswer>no answer) yet for icmp_seq=(?P<seq>\d+)$~", $chunk, $matches);
                        if (! empty($matches[0])) {
                            $matches["ip"][0] = $ip;
                            $matches["startTime"][0] = $this->startTime[$ip];
                            $matches = array_filter($matches, "is_string", ARRAY_FILTER_USE_KEY);
                            $matches = array_map("current", $matches);
                            $matches = array_filter($matches);
                            $this->startTime[$ip] = time();
                            $callable(new PingStateBad($matches));
                        }
                    });
                }

                if ($process->stderr !== null) {
                    $process->stderr->on('data', function ($chunk) use ($ip, $callable) {
                        preg_match_all("~connect: (?P<error>.+)$~U", $chunk, $matches);
                        $matches["ip"][0] = $ip;
                        $matches["startTime"][0] = $this->startTime[$ip];
                        $matches = array_filter($matches, "is_string", ARRAY_FILTER_USE_KEY);
                        $matches = array_map("current", $matches);
                        $matches = array_filter($matches);
                        $callable(new PingStateError($matches));
                    });
                }

                $process->on('exit', function ($code) use ($ip, $callable, $resolved) {
                    unset($this->state[$ip]);
                    // Process end : restart it after 1s to have an answer with the other one
                    // Arrive when IPv6 routing is not available at start.
                    \React\Promise\Timer\sleep(1.0)->then(function () use ($resolved, $callable) {
                        $this->run($resolved, $callable);
                    });
                });
            }
        }
    }

    /**
     * Try to found the ping command in system
     *
     * @return string
     */
    public function pingCmdInPath(): string
    {
        $path = getenv("PATH");
        if (! is_string($path)) {
            $path = "/bin";
        }
        $pathes = explode(":", $path);
        $cmd = "";
        foreach ($pathes as $path) {
            if (! file_exists("$path/ping") || ! is_executable("$path/ping")) {
                continue;
            }
            $cmd = "$path/ping";
            break;
        }
        if ($cmd === "") {
            throw new \Exception("Can not find the command 'ping' in PATH");
        }
        return $cmd;
    }
}
