<?php

/**
 * @author Dominique Fournier <dominique@fournier38.fr>
 * @license BSD
 */

declare(strict_types=1);

namespace App\Service;

use React\Async;
use React\Dns\Config\Config;
use React\Dns\Model\Message;
use React\Dns\Resolver\Factory;
use React\Promise;

class Dns
{
    private bool $ipv4 = true;
    private bool $ipv6 = true;

    public function ipv4(bool $val): self
    {
        $this->ipv4 = $val;

        return $this;
    }

    public function ipv6(bool $val): self
    {
        $this->ipv6 = $val;

        return $this;
    }

    /**
     * @param array<string> $dns The entry to resolve
     * @return array<string, array<string>> of entries with array of IPS
     */
    public function directDNS(array $dns): array
    {
        $config = Config::loadSystemConfigBlocking();
        if (empty($config->nameservers)) {
            throw new \Exception("DNS : Can not get NameServer from /etc/resolv.conf");
        }

        $factory = new Factory();
        $resolver = $factory->create($config);

        $promises = [];
        $ips = [];
        foreach ($dns as $entry) {
            $ips[$entry] = [];
            if ($this->ipv4) {
                if (filter_var($entry, FILTER_VALIDATE_IP, FILTER_FLAG_IPV4) !== false) {
                    $ips[$entry] = array_merge($ips[$entry], [$entry]);
                } else {
                    $promises["IPv4" . $entry] = $resolver->resolveAll($entry, Message::TYPE_A)->then(
                        function (array $resolved) use (&$ips, $entry) {
                            $ips[$entry] = array_merge($ips[$entry], $resolved);
                        },
                        function (\Throwable $e) {
                        }
                    );
                }
            }
            if ($this->ipv6) {
                if (filter_var($entry, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6) !== false) {
                    $ips[$entry] = array_merge($ips[$entry], [$entry]);
                } else {
                    $promises["IPv6" . $entry] = $resolver->resolveAll($entry, Message::TYPE_AAAA)->then(
                        function (array $resolved) use (&$ips, $entry) {
                            $ips[$entry] = array_merge($ips[$entry], $resolved);
                        },
                        function (\Throwable $e) {
                        }
                    );
                }
            }
        }

        // Wait the end of all the requests
        Async\await(Promise\all($promises));
        return $ips;
    }
}
