<?php

/**
 * @author Dominique Fournier <dominique@fournier38.fr>
 * @license BSD
 */

declare(strict_types=1);

namespace App\Service;

use GetOpt\ArgumentException;
use GetOpt\GetOpt;
use GetOpt\ArgumentException\Missing;
use GetOpt\Option;

/**
 * Manage the options of the command line
 */
class Getopts
{
    /**
     * @var array<\GetOpt\Operand>
     */
    private array $names = [];
    private bool $ipv4 = true;
    private bool $ipv6 = true;
    private string $display = "History";

    public function __construct()
    {
        $getOpt = new GetOpt();
        $displayFormat = ["RealTime", "History"];

        $getOpt->addOptions([
            Option::create("?", "help", GetOpt::NO_ARGUMENT)
                ->setDescription("Show this help and quit"),

            Option::create("4", null, GetOpt::NO_ARGUMENT)
                ->setDescription("Only IPv4 addresses"),

            Option::create("6", null, GetOpt::NO_ARGUMENT)
                ->setDescription("Only IPv6 addresses"),

            Option::create("d", "display", GetOpt::REQUIRED_ARGUMENT)
                ->setDescription("Display format : " . implode(",", $displayFormat) . ")"),
        ]);

        // process arguments and catch user errors
        try {
            try {
                $getOpt->process();
            } catch (Missing $exception) {
                // catch missing exceptions if help is requested
                if (!$getOpt->getOption('help')) {
                    throw $exception;
                }
            }
            if ($getOpt->getOption("4") && $getOpt->getOption("6")) {
                throw new ArgumentException("Can not have -4 and -6 option simultaneously");
            }
            if (empty($getOpt->getOperands())) {
                throw new ArgumentException("Missing required destination address");
            }
            $this->display = strval($getOpt->getOption("d") ?:  $getOpt->getOption("display") ?: $this->display);
            if (! in_array($this->display, $displayFormat)) {
                throw new ArgumentException(
                    "Unknown display format. Must be in : " . implode(",", $displayFormat)
                );
            }
        } catch (ArgumentException $exception) {
            file_put_contents("php://stderr", "ERROR : " . $exception->getMessage() . PHP_EOL);
            echo PHP_EOL . $getOpt->getHelpText();
            exit;
        }

        // show help and quit
        if ($getOpt->getOption("help")) {
            echo $getOpt->getHelpText();
            exit;
        }

        if ($getOpt->getOption("4")) {
            $this->ipv6 = false;
        }
        if ($getOpt->getOption("6")) {
            $this->ipv4 = false;
        }

        $this->names = $getOpt->getOperands();
    }

    /**
     * Return the provided names
     *
     * @return array<\GetOpt\Operand>
     */
    public function getNames(): array
    {
        return $this->names;
    }

    /**
     * Return the IPv4 state
     */
    public function getIPv4(): bool
    {
        return $this->ipv4;
    }

    /**
     * Return the IPv6 state
     */
    public function getIPv6(): bool
    {
        return $this->ipv6;
    }

    /**
     * Return the Display Format
     */
    public function getDisplay(): string
    {
        return $this->display;
    }
}
